/* 
 *  Lsystem version 1.0, Copyright (C) 1994 G. Hutchings
 *  Lsystem comes with ABSOLUTELY NO WARRANTY.
 *  This is free software, and you are welcome to redistribute it
 *  under certain conditions; see the file COPYING for details.
 */

/* Lsystem output conversion program */

#include <stdio.h>
#include <time.h>
#include "ls2dev.h"

FILE *ifp = stdin;		 /* input stream */
FILE *ofp = stdout;		 /* output stream */

int lang = -1;			 /* output language */
int num_cmds = 0;		 /* no. of commands read */
int max_cmds = MAX_CMDS;	 /* max. no. of commands */

struct cmd *cmdlist = NULL;	 /* command list */

double xmin, xmax, ymin, ymax;	 /* bounding box of drawing */

double linewidth = 1.0;		 /* PostScript line width */

int
main(int argc, char *argv[])
{
    extern int optind, opterr;
    extern char *optarg;
    char line[BUFSIZ], infile[BUFSIZ], cmd;
    int c, i, line_num = 0;
    double arg1, arg2, arg3;

    /* Parse arguments */
    while ((c = getopt(argc, argv, "l:o:L:")) != EOF) {
	switch (c) {
	case 'l':
	    linewidth = atof(optarg);
	    if (linewidth <= 0.0) fatal("invalid line width");
	    break;
	case 'o':
	    ofp = fopen(optarg, "w");
	    if (ofp == NULL) {
		fprintf(stderr, "Can't open %s\n", optarg);
		return 1;
	    }
	    break;
	case 'L':
	    for (i = 0; languages[i].name != NULL; i++) {
		if (!strcmp(optarg, languages[i].name)) lang = i;
	    }

	    if (lang < 0) {
		fprintf(stderr, "Language `%s' is not recognized\n", optarg);
		return 1;
	    }
	    break;
	default:
	    usage();
	}
    }

    if (lang < 0) fatal("language not specified (use -L option)");

    /* Last argument (if any) is input file */
    if (optind != argc) {
	ifp = fopen(argv[optind], "r");
	if (ifp == NULL) {
	    fprintf(stderr, "Can't open %s\n", argv[optind]);
	    exit(1);
	}
	strcpy(infile, argv[optind]);
    } else {
	strcpy(infile, "<stdin>");
    }

    /* Allocate command list */
    cmdlist = ALLOC(struct cmd, max_cmds);
    if (cmdlist == NULL) fatal("out of memory");

    /* Read input line by line */
    while (fgets(line, BUFSIZ, ifp) != NULL) {
	line_num++;

	/* Parse command */
	switch (line[0]) {
	case 'M':
	case 'L':
	    if (sscanf(line, "%c %lf %lf", &cmd, &arg1, &arg2) != 3)
		error(infile, line_num, "invalid command format");
	    arg3 = 0.0;
	    if (num_cmds == 0) {
		xmin = xmax = arg1;
		ymin = ymax = arg2;
	    } else {
		xmin = MIN(xmin, arg1);
		xmax = MAX(xmax, arg1);
		ymin = MIN(ymin, arg2);
		ymax = MAX(ymax, arg2);
	    }
	    break;
	case 'C':
	    if (sscanf(line, "%c %lf %lf %lf", &cmd, &arg1, &arg2, &arg3) != 4)
		error(infile, line_num, "invalid command format");
	    break;
	default:
	    error(infile, line_num, "unrecognized command");
	}

	/* Add it to list */
	cmdlist[num_cmds].name = cmd;
	cmdlist[num_cmds].arg1 = arg1;
	cmdlist[num_cmds].arg2 = arg2;
	cmdlist[num_cmds].arg3 = arg3;
	num_cmds++;

	/* Increase list size if required */
	if (num_cmds == max_cmds) {
	    max_cmds *= 2;
	    cmdlist = REALLOC(cmdlist, struct cmd, max_cmds);
	    if (cmdlist == NULL) fatal("out of memory");
	}
    }

    /* Invoke drawing function */
    (languages[lang].func)();

    /* Finish */
    if (ifp != stdin) fclose(ifp);
    if (ofp != stdout) fclose(ofp);

    return 0;
}

void
do_postscript()
{
    int i, line = 0, nl = 0;

    double x, y, r, g, b, fac;
    double xmid = (xmin+xmax)/2;
    double ymid = (ymin+ymax)/2;
    double xlen = (xmax-xmin);
    double ylen = (ymax-ymin);

    time_t tloc = time(NULL);
    char *timestring = ctime(&tloc);

    /* Calculate scaling factor */
    if (xlen == 0.0 || ylen == 0.0) return;
    fac = 0.8*MIN(PS_XLEN/xlen, PS_YLEN/ylen);

    /* Print header */
    fprintf(ofp, "%%\!PS-Adobe-1.0\n");
    fprintf(ofp, "%%%%Title: L System fractal picture\n");
    fprintf(ofp, "%%%%Creator: ls2dev\n");
    fprintf(ofp, "%%%%CreationDate: %s", timestring);
    fprintf(ofp, "%%%%Pages: 1\n");
    fprintf(ofp, "%%%%EndComments\n");

    /* Print drawing definitions */
    fprintf(ofp, "/N { newpath } bind def\n");
    fprintf(ofp, "/M { moveto } bind def\n");
    fprintf(ofp, "/L { lineto } bind def\n");
    fprintf(ofp, "/S { stroke } bind def\n");
    fprintf(ofp, "/C { setrgbcolor } bind def\n");

    fprintf(ofp, "%%%%EndProlog\n");
    fprintf(ofp, "%%%%Page: ? 1\n");

    fprintf(ofp, "%.3f setlinewidth\n", linewidth);

    /* Do commands */
    for (i = 0; i < num_cmds; i++) {
	switch (cmdlist[i].name) {
	case 'M':
	    x = cmdlist[i].arg1;
	    y = cmdlist[i].arg2;
	    x = (x-xmid)*fac+(PS_XLEN/2);
	    y = (y-ymid)*fac+(PS_YLEN/2);
	    if (line) fprintf(ofp, "S ");
	    fprintf(ofp, "N %.1f %.1f M\n", x, y);
	    line = 1;
	    nl = 0;
	    break;
	case 'L':
	    x = cmdlist[i].arg1;
	    y = cmdlist[i].arg2;
	    x = (x-xmid)*fac+(PS_XLEN/2);
	    y = (y-ymid)*fac+(PS_YLEN/2);
	    fprintf(ofp, "%.1f %.1f L\n", x, y);
	    if (nl++ == PS_FLUSH) {
		nl = 0;
		fprintf(ofp, "S N %.1f %.1f M\n", x, y);
	    }
	    break;
	case 'C':
	    r = cmdlist[i].arg1;
	    g = cmdlist[i].arg2;
	    b = cmdlist[i].arg3;
	    if (line) fprintf(ofp, "S ");
	    fprintf(ofp, "%.3f %.3f %.3f C\n", r, g, b);
	    fprintf(ofp, "N %.1f %.1f M\n", x, y);
	    break;
	}
    }

    /* Finish line (if any) */
    if (line) fprintf(ofp, "S\n");

    /* Print trailer */
    fprintf(ofp, "showpage\n");
    fprintf(ofp, "%%%%Trailer\n");
}

void
do_gnuplot()
{
    int i;

    for (i = 0; i < num_cmds; i++) {
	switch (cmdlist[i].name) {
	case 'M':
	    fprintf(ofp, "\n");
	    /* FALLTHROUGH */
	case 'L':
	    fprintf(ofp, "%.3f %.3f\n", cmdlist[i].arg1, cmdlist[i].arg2);
	    break;
	case 'C':		 /* ignored in gnuplot */
	    break;
	}
    }
}

void
error(char *file, int line, char *msg)
{
    fprintf(stderr, "ls2dev: Error: %s, line %d: %s\n", file, line, msg);
    exit(1);
}

void
fatal(char *msg)
{
    fprintf(stderr, "ls2dev: Fatal: %s\n", msg);
    exit(1);
}

void
usage()
{
    fprintf(stderr, "Usage: ls2dev [-L lang] [-o file] [language-options] file\n\n");
    fprintf(stderr, "-l width\tset PostScript line width\n");
    exit(1);
}
