/* 
 *  Lsystem version 1.1, Copyright (C) 1994 G. Hutchings
 *  Lsystem comes with ABSOLUTELY NO WARRANTY.
 *  This is free software, and you are welcome to redistribute it
 *  under certain conditions; see the file COPYING for details.
 */

/* Rule/macro/colour routines */

#include <stdio.h>
#include "rule.h"
#include "defs.h"

struct rule *toprule = NULL;	 /* top rule in rule list */
struct macro *topmacro = NULL;	 /* top macro in macro list */
struct colour *topcolour = NULL; /* top colour in colour list */

char *startstring;		 /* initial string */

double angle_inc;		 /* angle increment */

int initialized = 0;		 /* initialization flag */
int num_iterations;		 /* no. of iterations */

/* Set initial lsystem parameters */
void
initialize(char *str, int iterations, double angle)
{
    static double pi = 3.1415926535;

    if (iterations < 0) yyerror("invalid no. of iterations");

    if (!initialized) {
	initialized = 1;
	angle_inc = (angle*pi)/180.0;
	num_iterations = iterations;
	startstring = ALLOC(char, strlen(str)+1);
	if (startstring == NULL) MEMERROR;
	strcpy(startstring, str);
    } else {
	yyerror("only one initialization allowed");
    }
}

/* Add a macro to the macro list */
void
addmacro(char name, char *def)
{
    struct macro *m, *new;
    extern char errbuf[];

    for (m = topmacro; m != NULL; m = m->next) {
	if (m->name == name) {
	    sprintf(errbuf, "macro already defined: %c", name);
	    yyerror(errbuf);
	    return;
	}
    }

    new = ALLOC(struct macro, 1);
    if (new == NULL) MEMERROR;
    new->name = name;
    new->def = ALLOC(char, strlen(def)+1);
    if (new->def == NULL) MEMERROR;
    strcpy(new->def, def);
    new->next = NULL;

    if (topmacro == NULL) {
	topmacro = new;
    } else {
	for (m = topmacro; m->next != NULL; m = m->next);
	m->next = new;
    }
}

/* Add a rule to the rule list */
void
addrule(char name, char *def)
{
    struct rule *r, *new;
    extern char errbuf[];

    for (r = toprule; r != NULL; r = r->next) {
	if (r->name == name) {
	    sprintf(errbuf, "rule already defined: %c", name);
	    yyerror(errbuf);
	    return;
	}
    }

    new = ALLOC(struct rule, 1);
    if (new == NULL) MEMERROR;
    new->name = name;
    new->def = ALLOC(char, strlen(def)+1);
    if (new->def == NULL) MEMERROR;
    strcpy(new->def, def);
    new->next = NULL;

    if (toprule == NULL) {
	toprule = new;
    } else {
	for (r = toprule; r->next != NULL; r = r->next);
	r->next = new;
    }
}

/* Add a colour to the colour list */
void
addcolour(char name, double red, double green, double blue)
{
    struct colour *c, *new;
    extern char errbuf[];

    if (red < 0.0 || red > 1.0) yyerror("invalid colour value");
    if (green < 0.0 || green > 1.0) yyerror("invalid colour value");
    if (blue < 0.0 || blue > 1.0) yyerror("invalid colour value");

    for (c = topcolour; c != NULL; c = c->next) {
	if (c->name == name) {
	    sprintf(errbuf, "colour already defined: %c", name);
	    yyerror(errbuf);
	    return;
	}
    }

    new = ALLOC(struct colour, 1);
    if (new == NULL) MEMERROR;
    new->name = name;
    new->red = red;
    new->green = green;
    new->blue = blue;
    new->next = NULL;

    if (topcolour == NULL) {
	topcolour = new;
    } else {
	for (c = topcolour; c->next != NULL; c = c->next);
	c->next = new;
    }
}

/* Return the value of a macro */
char *
expandmacro(char name)
{
    struct macro *m;

    for (m = topmacro; m != NULL; m = m->next) {
	if (m->name == name) return m->def;
    }

    return NULL;
}

/* Return the RGB values of a colour (if defined) */
int
findcolour(char name, double *red, double *green, double *blue)
{
    struct colour *c;

    for (c = topcolour; c != NULL; c = c->next) {
	if (c->name == name) {
	    *red = c->red;
	    *green = c->green;
	    *blue = c->blue;
	    return 1;
	}
    }

    return 0;
}

/* Iterate initial string and return result */
char *
iterate()
{
    char *str1 = startstring, *str2, *rp;
    long size = MAX_STR_LEN;
    extern int stringinfo;
    struct rule *r;
    int s1, s2, i;

    for (i = 1; i <= num_iterations; i++) {
	/* Initialize new string */
	str2 = ALLOC(char, size);
	if (str2 == NULL) MEMERROR;
	s2 = 0;

	/* Scan current string */
	for (s1 = 0; str1[s1] != NULL; s1++) {
	    /* Find applicable rule (if any) */
	    for (r = toprule; r != NULL; r = r->next) {
		if (r->name == str1[s1]) break;
	    }

	    if (r == NULL) {
		/* No rule - copy character */
		str2[s2++] = str1[s1];
		if (s2 == size) {
		    size *= 2;
		    str2 = REALLOC(str2, char, size);
		    if (str2 == NULL) MEMERROR;
		}
	    } else {
		/* Found a rule - expand its definition */
		for (rp = r->def; *rp != NULL; rp++) {
		    str2[s2++] = *rp;
		    if (s2 == size) {
			size *= 2;
			str2 = REALLOC(str2, char, size);
			if (str2 == NULL) MEMERROR;
		    }
		}
	    }
	}

	/* Terminate string */
	str2[s2] = NULL;

	/* Destroy old string */
	free(str1);
	str1 = str2;

	/* Print string info if required */
	if (stringinfo) fprintf(stderr, "Iteration: %3d  String size: %d\n",
				i, strlen(str1));
    }

    return str1;
}
